<?php
/**
 * The Ingo_Script:: class provides a common abstracted interface to the
 * script-generation subclasses.
 *
 * $Horde: ingo/lib/Script.php,v 1.5 2003/04/25 23:40:58 slusarz Exp $
 *
 * @author  Brent J. Nordquist <bjn@horde.org>
 * @version $Revision: 1.5 $
 * @since   Ingo 0.1
 * @package ingo
 */
class Ingo_Script {

    /**
     * The name of the current script type.
     *
     * @var string $_type
     */
    var $_type = '';

    /**
     * The script class' additional parameters.
     *
     * @var array $_params
     */
    var $_params = array();

    /**
     * The list of actions allowed (implemented) for this script type.
     *
     * @var array $_actions
     */
    var $_actions = array();

    /**
     * The list of tests allowed (implemented) for this script type.
     *
     * @var array $_tests
     */
    var $_tests = array();

    /**
     * Attempts to return a concrete Ingo_Script instance based on $script.
     *
     * @param string $script          The type of Ingo_Script subclass to
     *                                return. The code is dynamically
     *                                included.
     * @param optional array $params  Hash containing additional paramters to
     *                                be passed to the subclass' constructor.
     *
     * @return object Ingo_Script  The newly created concrete Ingo_Script
     *                             instance, or false on error.
     */
    function &factory($script, $params = array())
    {
        $script = basename($script);
        require_once dirname(__FILE__) . '/Script/' . $script . '.php';
        $class = 'Ingo_Script_' . $script;
        if (class_exists($class)) {
            return new $class($params);
        } else {
            Horde::fatal(PEAR::raiseError(sprintf(_("Unable to load the definition of %s."), $class)), __FILE__, __LINE__);
        }
    }

    /**
     * Attempts to return a reference to a concrete Ingo_Script instance
     * based on $script. It will only create a new instance if no
     * Ingo_Script instance with the same parameters currently exists.
     *
     * This method must be invoked as: $var = &Ingo_Script::singleton()
     *
     * @param string $script          The type of concrete Ingo_Script
     *                                subclass to return. This is based on the
     *                                script type ($script). The code is
     *                                dynamically included.
     * @param optional array $params  A hash containing additional parameters
     *                                for the subclass.
     *
     * @return object Ingo_Script  The concrete Ingo_Script reference, or
     *                             false on error.
     */
    function &singleton($script, $params = array())
    {
        static $instances;

        if (!isset($instances)) {
            $instances = array();
        }

        $signature = serialize(array($script, $params));
        if (!isset($instances[$signature])) {
            $instances[$signature] = &Ingo_Script::factory($script, $params);
        }

        return $instances[$signature];
    }

    /**
     * Returns the name of the current script type.
     *
     * @return  String containing the name of the current script type.
     */
    function getType()
    {
        return $this->type;
    }

    /**
     * Returns the current script class' additional parameters.
     *
     * @return  Hash containing the script class' additional parameters.
     */
    function getParams()
    {
        return $this->params;
    }

    /**
     * Returns the available actions for this script type.
     *
     * @access public
     *
     * @return array  The list of available actions.
     */
    function availableActions()
    {
        return $this->_actions;
    }

    /**
     * Returns the available tests for this script type.
     *
     * @access public
     *
     * @return array  The list of tests actions.
     */
    function availableTests()
    {
        return $this->_tests;
    }

    /**
     * Returns a script previously generated with generate().
     *
     * @abstract
     *
     * @access public
     *
     * @return string  The script.
     */
    function toCode()
    {
        return '';
    }

    /**
     * Generates the script to do the filtering specified in
     * the rules.
     *
     * @abstract
     *
     * @access public
     *
     * @return string  The script.
     */
    function generate()
    {
        return '';
    }

    /**
     * Perform the filtering specified in the rules.
     *
     * @abstract
     *
     * @access public
     *
     * @return boolean  True if filtering performed, false if not.
     */
    function perform()
    {
        return false;
    }

}
