<?php
/**
 * The Ingo_Script_procmail:: class represents a Procmail script generator.
 *
 * Copyright 2003-2005 Brent J. Nordquist <bjn@horde.org>
 *
 * See the enclosed file LICENSE for license information. If you
 * did not receive this file, see http://www.horde.org/licenses.
 *
 * $Horde: ingo/lib/Script/procmail.php,v 1.9.2.2 2005/01/03 11:25:56 jan Exp $
 *
 * @author  Brent J. Nordquist <bjn@horde.org>
 * @version $Revision: 1.9.2.2 $
 * @since   Ingo 0.1
 * @package ingo
 */
class Ingo_Script_procmail extends Ingo_Script {

    /**
     * The recipes that make up the code.
     *
     * @var array $_recipes
     */
    var $_recipes = array();

    /**
     * Constructs a new procmail script generator object.
     *
     * @access public
     *
     * @param array $params   A hash containing script parameters.
     */
    function Ingo_Script_procmail($params)
    {
        $this->_type = 'procmail';
        $this->_params = $params;
        $this->_actions = array(INGO_STORAGE_ACTION_KEEP,
                                INGO_STORAGE_ACTION_MOVE,
                                INGO_STORAGE_ACTION_DISCARD);
        $this->_tests = array('contains', 'not contain', 'is', 'not is',
                              'begins with', 'not begins with', 'ends with',
                              'not ends with', 'exists', 'not exist', 
                              'less than', 'greater than');
    }

    /**
     * Returns a script previously generated with generate().
     *
     * @access public
     *
     * @return string  The procmail script.
     */
    function toCode()
    {
        $code = '';
        foreach ($this->_recipes as $recipe) {
            $code .= $recipe . "\n\n";
        }
        return $code;
    }

    /**
     * Generates the procmail script to do the filtering specified in
     * the rules.
     *
     * @access public
     *
     * @return string  The procmail script.
     */
    function generate()
    {
        global $ingo;

        /* Initialization */
        $_recipes = array();
        $condition = &new Procmail_Condition();

        /* Rules */
        $rules = $ingo['storage']->retrieve('rules');
        foreach ($rules as $rule_number => $rule) {
            if (empty($rule['action'])) {
                continue;
            }
            switch ($rule['action']) {
            case INGO_STORAGE_ACTION_BLACKLIST:
                /* Prepare action (same for all blacklist recipes) */
                $action = &new Procmail_Action();
                $folder = $ingo['storage']->retrieve('blacklist_folder');
                if (empty($folder)) {
                    $action->discard();
                } elseif ($folder == INGO_BLACKLIST_MARKER) {
                    // FIXME
                    $action->discard();
                } else {
                    $action->deliverToFolder($folder);
                }

                /* Blacklist recipes */
                $blacklist = $ingo['storage']->retrieve('blacklist');
                if (!empty($blacklist)) {
                    $this->addComment(_("Blacklisted Addresses"));
                    foreach ($blacklist as $address) {
                        if (!empty($address)) {
                            $condition->from($address);
                            $this->addRecipe($condition, $action);
                        }
                    }
                }
                break;

            case INGO_STORAGE_ACTION_WHITELIST:
                /* Prepare action (same for all whitelist recipes) */
                $action = &new Procmail_Action();
                $action->keep();

                /* Whitelist recipes */
                $whitelist = $ingo['storage']->retrieve('whitelist');
                if (!empty($whitelist)) {
                    $this->addComment(_("Whitelisted Addresses"));
                    foreach ($whitelist as $address) {
                        if (!empty($address)) {
                            $condition->from($address);
                            $this->addRecipe($condition, $action);
                        }
                    }
                }
                break;

            case INGO_STORAGE_ACTION_VACATION:
            case INGO_STORAGE_ACTION_MOVE:
            case INGO_STORAGE_ACTION_KEEP:
            case INGO_STORAGE_ACTION_DISCARD:
            case INGO_STORAGE_ACTION_REJECT:
            case INGO_STORAGE_ACTION_REDIRECT:
            case INGO_STORAGE_ACTION_REDIRECTKEEP:
            default:
                $this->addComment(sprintf(_("Unsupported Rule: %s"), $rule['action']));
            }
        }
        return $this->toCode();
    }

    /**
     * Add a comment to the procmail script.
     *
     * @access public
     *
     * @param string $comment  The comment to add.
     */
    function addComment($comment)
    {
        $this->_recipes[] = "# $comment";
    }

    /**
     * Add a recipe to the procmail script.
     *
     * @access public
     *
     * @param object Procmail_Condition $condition  The condition for the
     *                                              recipe.
     * @param object Procmail_Action $action        The action for the recipe.
     */
    function addRecipe($condition, $action)
    {
        $recipe = ":0 " . $condition->getFlags() . $action->getFlags() . $action->getLock() . "\n";
        $recipe .= $condition->getCondition()."\n";
        $recipe .= $action->getAction();
        $this->_recipes[] = $recipe;
    }

}

/**
 * The Procmail_Condition:: class represents a Procmail condition.
 *
 * @author  Brent J. Nordquist <bjn@horde.org>
 * @since   Ingo 0.1
 * @package ingo
 */
class Procmail_Condition {

    var $_condition = '';
    var $_flags = '';

    /**
     * Constructs a new procmail script generator object.
     *
     * @access public
     */
    function Procmail_Condition()
    {
    }

    /**
     * Returns the recipe condition line (starting with "* ", but not
     * including newline).
     *
     * @access public
     *
     * @return string  The recipe condition line.
     */
    function getCondition()
    {
        return $this->_condition;
    }

    /**
     * Returns the recipe flags associated with the condition.
     *
     * @access public
     *
     * @return string  The recipe flags.
     */
    function getFlags()
    {
        return $this->_flags;
    }

    /**
     * Set a condition that matches email "From:" the given address.
     *
     * @access public
     *
     * @param string  The From: email address to match.
     */
    function from($address)
    {
        $this->_condition = "* ^From:.*$address";
        $this->_flags = '';
    }

}

/**
 * The Procmail_Action:: class represents a Procmail action.
 *
 * @author  Brent J. Nordquist <bjn@horde.org>
 * @since   Ingo 0.1
 * @package ingo
 */
class Procmail_Action {

    var $_action = '';
    var $_flags = '';
    var $_lock = '';

    /**
     * Constructs a new procmail script generator object.
     *
     * @access public
     */
    function Procmail_Action()
    {
    }

    /**
     * Returns the recipe action line (not including newline).
     *
     * @access public
     *
     * @return string  The recipe action line.
     */
    function getAction()
    {
        return $this->_action;
    }

    /**
     * Returns the recipe flags associated with the action.
     *
     * @access public
     *
     * @return string  The recipe flags.
     */
    function getFlags()
    {
        return $this->_flags;
    }

    /**
     * Returns the lock flag (if applicable to this action).
     *
     * @access public
     *
     * @return string  The lock flag.
     */
    function getLock()
    {
        return $this->_lock;
    }

    /**
     * Set an action to keep the email.
     *
     * @access public
     */
    function keep()
    {
        $this->_action = '$DEFAULT';  // NB: literal $
        $this->_flags = '';
        $this->_lock = ':';
    }

    /**
     * Set an action to discard the email.
     *
     * @access public
     */
    function discard()
    {
        $this->_action = "/dev/null";
        $this->_flags = '';
        $this->_lock = '';
    }

    /**
     * Returns a procmail-ready mailbox path, converting IMAP folder
     * pathname conventions as necessary.
     *
     * @access public
     *
     * @param string  The IMAP folder name.
     *
     * @return string  The procmail mailbox path.
     */
    function procmailPath($folder)
    {
        global $conf;

        // NB: where $DEFAULT appears here, that's a literal $
        // (not a PHP variable)
        if ($conf['script']['path_style'] == 'maildir') {
            if (empty($folder) || $folder == 'INBOX') {
                return '$DEFAULT';
            }
            if (substr($folder, 0, 6) == 'INBOX.') {
                $folder = substr($folder, 6);
            }
            return '$DEFAULT/.'.$folder.'/';
        } else {
            if (empty($folder) || $folder == 'INBOX') {
                return '$DEFAULT';
            }
            return $folder;
        }
    }

    /**
     * Set an action to deliver the email to the given folder.
     *
     * @access public
     *
     * @param string  The folder to deliver the email to.
     */
    function deliverToFolder($folder)
    {
        $this->_action = $this->procmailPath($folder);
        $this->_flags = '';
        $this->_lock = ':';
    }

}
