<?php
/**
 * The Ingo_Script_sieve:: class represents a Sieve Script.
 *
 * $Horde: ingo/lib/Script/sieve.php,v 1.13.2.2 2004/09/22 17:43:46 slusarz Exp $
 *
 * @author  Mike Cochrane <mike@graftonhall.co.nz>
 * @version $Revision: 1.13.2.2 $
 * @since   Ingo 0.1
 * @package ingo
 */
class Ingo_Script_sieve extends Ingo_Script {

    /**
     * The blocks that make up the code.
     *
     * @var array $_blocks
     */
    var $_blocks = array();

    /**
     * Constructs a new Sieve script generator object.
     *
     * @access public
     *
     * @param optional array $params  A hash containing script parameters.
     */
    function Ingo_Script_sieve($params = array())
    {
        $this->_params = $params;
        $this->_type = 'sieve';
        $this->_actions = array(INGO_STORAGE_ACTION_KEEP,
                                INGO_STORAGE_ACTION_MOVE,
                                INGO_STORAGE_ACTION_DISCARD,
                                INGO_STORAGE_ACTION_REDIRECT,
                                INGO_STORAGE_ACTION_REDIRECTKEEP,
                                INGO_STORAGE_ACTION_REJECT);
        $this->_tests = array('contains', 'not contain', 'is', 'not is',
                              'begins with', 'not begins with', 'ends with',
                              'not ends with', 'exists', 'not exist',
                              'less than', 'greater than');
    }

    /**
     * Returns a script previously generated with generate().
     *
     * @access public
     *
     * @return string  The Sieve script.
     */
    function toCode()
    {
        $code = '';
        $requires = $this->requires();

        if (count($requires) > 1) {
            $stringlist = '';
            foreach ($this->requires() as $require) {
                $stringlist .= (empty($stringlist)) ? '"' : ', "';
                $stringlist .= $require . '"';
            }
            $code .= 'require [' . $stringlist . '];' . "\n\n";
        } elseif (count($requires) == 1) {
            foreach ($this->requires() as $require) {
                $code .= 'require "' . $require . '";' . "\n\n";
            }
        }

        foreach ($this->_blocks as $block) {
            $code .= $block->toCode() . "\n";
        }

        return $code;
    }

    /**
     * Escape a string according to Sieve RFC 3028 [2.4.2].
     *
     * @access public
     *
     * @param string $string  The string to escape.
     *
     * @return string  The escaped string.
     */
    function escapeString($string)
    {
        $string = str_replace('\\', addslashes('\\'), $string);
        $string = str_replace('"', addslashes('"'), $string);

        return $string;
    }

    function addBlock($block)
    {
        $this->_blocks[] = $block;
    }

    function check()
    {
        foreach ($this->_blocks as $block) {
            $res = $block->check();
            if ($res !== true) {
                return $res;
            }
        }

        return true;
    }

    function requires()
    {
        $requires = array();

        foreach ($this->_blocks as $block) {
            $requires = array_merge($requires, $block->requires());
        }

        $requires = array_unique($requires);

        return $requires;
    }

    /**
     * Generates the Sieve script to do the filtering specified in
     * the rules.
     *
     * @return string  The Sieve script.
     */
    function generate()
    {
        global $ingo;

        $_blocks = array();

        /* Blacklist Script */
        $blacklist = $ingo['storage']->retrieve('blacklist');
        $blacklistBlocks = array();;
        if (!empty($blacklist)) {
            $folder = $ingo['storage']->retrieve('blacklist_folder');
            $action = array();

            if (empty($folder)) {
                $action[] = &new Sieve_Action_Discard();
            } elseif ($folder == INGO_BLACKLIST_MARKER) {
                $action[] = &new Sieve_Action_Addflag(array('flags' => array('\\Deleted')));
                $action[] = &new Sieve_Action_Keep();
                $action[] = &new Sieve_Action_Removeflag(array('flags' => array('\\Deleted')));
            } else {
                $action[] = &new Sieve_Action_Addflag(array('flags' => array('\\Seen')));
                $action[] = &new Sieve_Action_Fileinto(array('folder' => $folder));
                $action[] = &new Sieve_Action_Removeflag(array('flags' => array('\\Seen')));
            }

            $action[] = &new Sieve_Action_Stop();

            $comment = &new Sieve_Comment(_("Blacklisted Addresses"));
            $blacklistBlocks[] = $comment;

            // split the test up to only do 5 addresses at a time.
            $temp = array();
            $wildcards = array();
            foreach ($blacklist as $address) {
                if (!empty($address)) {
                    if (strstr($address, '*') !== false || strstr($address, '?') !== false) {
                        $wildcards[] = $address;
                    } else {
                        $temp[] = $address;
                    }
                }
                if (count($temp) == 5) {
                    $test = &new Sieve_Test_Address(array('headers' => "From\nSender\nResent-From", 'addresses' => implode("\n", $temp)));
                    $if = &new Sieve_If($test);
                    $if->setActions($action);
                    $this->addBlock($if);
                    $temp = array();
                }
                if (count($wildcards) == 5) {
                    $test = &new Sieve_Test_Address(array('headers' => "From\nSender\nResent-From", 'match-type' => ':matches', 'addresses' => implode("\n", $wildcards)));
                    $if = &new Sieve_If($test);
                    $if->setActions($action);
                    $this->addBlock($if);
                    $wildcards = array();
                }
            }

            if (count($temp) > 0) {
                $test = &new Sieve_Test_Address(array('headers' => "From\nSender\nResent-From", 'addresses' => implode("\n", $temp)));
                $if = &new Sieve_If($test);
                $if->setActions($action);
                $blacklistBlocks[] = $if;
            }

            if (count($wildcards) > 0) {
                $test = &new Sieve_Test_Address(array('headers' => "From\nSender\nResent-From", 'match-type' => ':matches', 'addresses' => implode("\n", $wildcards)));
                $if = &new Sieve_If($test);
                $if->setActions($action);
                $blacklistBlocks[] = $if;
            }

        }

        /* Whitelist Script */
        $whitelist = $ingo['storage']->retrieve('whitelist');
        $whitelistBlocks = array();;
        if (!empty($whitelist)) {
            $whitelistBlocks[] = &new Sieve_Comment(_("Whitelisted Addresses"));
            $action = array();
            $action[] = &new Sieve_Action_Keep();
            $test = &new Sieve_Test_Address(array('headers' => "From\nSender\nResent-From", 'addresses' => implode("\n", $whitelist)));
            $if = &new Sieve_If($test);    
            $if->setActions($action);
            $whitelistBlocks[] = $if;
        }

        /* Vacation Script */
        $vacation = $ingo['storage']->retrieve('vacation');
        $vacationBlocks = array();
        if (array_key_exists('active', $vacation) && $vacation['active']) {
            $action = array();
            $vals = array('subject' => $vacation['subject'],
                          'days' => $vacation['days'],
                          'addresses' => $vacation['addresses'],
                          'reason' => $vacation['reason']);

            $action[] = &new Sieve_Action_Vacation($vals);

            $listHeaders = array();
            $listHeaders[] = 'List-Id';
            $listHeaders[] = 'List-Post';
            $listHeaders[] = 'List-Subscribe';
            $listHeaders[] = 'List-Unsubscribe';
            $listHeaders[] = 'List-Archive';
            $listHeaders[] = 'List-Help';
            $listHeaders[] = 'X-List-Name';

            $tests = array();
            if ($vacation['ignorelist']) {
                $tmp = &new Sieve_Test_Exists(array('headers' => implode("\n", $listHeaders)));
                $tests[] = &new Sieve_Test_Not($tmp);
            }


            $excludes = preg_split('(\n|\n|\r)', $vacation['excludes']);
            $addrs = array();
            foreach ($excludes as $addr) {
                $addr = trim($addr);
                if (!empty($addr)) {
                    $addrs[] = $addr;
                }
            }

            if (count($addrs) > 0) {
                $tmp = &new Sieve_Test_Address(array('headers' => "From\nSender\nResent-From", 'addresses' => implode("\n", $addrs)));
                $tests[] = &new Sieve_Test_Not($tmp);
            }

            $comment = &new Sieve_Comment(_("Vacation Message"));
            $vacationBlocks[] = $comment;

            if (count($tests) > 0) {
                $test = &new Sieve_Test_Allof($tests);
                $if = &new Sieve_If($test);
                $if->setActions($action);
                $vacationBlocks[] = $if;
            } else {
                $vacationBlocks[] = $action[0];
            }
        }

        $rule_number = 0;
        $filters = $ingo['storage']->retrieve('rules');
        foreach ($filters as $rule_number => $filter) {
            $rule_number++;

            $action = array();
            if (empty($filter['action'])) {
                continue;
            }
            switch ($filter['action']) {
                case INGO_STORAGE_ACTION_KEEP:
                    if (array_key_exists('flags', $filter) && is_array($filter['flags']) && count($filter['flags']) > 0) {
                        $action[] = &new Sieve_Action_Addflag(array('flags' => $filter['flags']));
                    }

                    $action[] = &new Sieve_Action_Keep();

                    if (array_key_exists('flags', $filter) && is_array($filter['flags']) && count($filter['flags']) > 0) {
                        $action[] = &new Sieve_Action_RemoveFlag(array('flags' => $filter['flags']));
                    }
                    break;

                case INGO_STORAGE_ACTION_DISCARD:
                    $action[] = &new Sieve_Action_Discard();
                    break;

                case INGO_STORAGE_ACTION_MOVE:
                    if (array_key_exists('flags', $filter) && is_array($filter['flags']) && count($filter['flags']) > 0) {
                        $action[] = &new Sieve_Action_Addflag(array('flags' => $filter['flags']));
                    }

                    $action[] = &new Sieve_Action_Fileinto(array('folder' => $filter['action-value']));

                    if (array_key_exists('flags', $filter) && is_array($filter['flags']) && count($filter['flags']) > 0) {
                        $action[] = &new Sieve_Action_RemoveFlag(array('flags' => $filter['flags']));
                    }
                    break;

                case INGO_STORAGE_ACTION_REJECT:
                    $action[] = &new Sieve_Action_Reject(array('reason' => $filter['action-value']));
                    break;

                case INGO_STORAGE_ACTION_REDIRECT:
                    $action[] = &new Sieve_Action_Redirect(array('address' => $filter['action-value']));
                    break;

                case INGO_STORAGE_ACTION_REDIRECTKEEP:
                    $action[] = &new Sieve_Action_Redirect(array('address' => $filter['action-value']));
                    $action[] = &new Sieve_Action_Keep();
                    break;

                case INGO_STORAGE_ACTION_WHITELIST:
                    foreach ($whitelistBlocks as $white) {
                        $this->addBlock($white);
                    }
                    continue 2;

                case INGO_STORAGE_ACTION_BLACKLIST:
                    foreach ($blacklistBlocks as $block) {
                        $this->addBlock($block);
                    }
                    continue 2;

                case INGO_STORAGE_ACTION_VACATION:
                    foreach ($vacationBlocks as $block) {
                        $this->addBlock($block);
                    }
                    continue 2;
            }

            $comment = &new Sieve_Comment($filter['name']);
            $this->addBlock($comment);

            if ($filter['stop']) {
                $action[] = &new Sieve_Action_Stop();
            }

            $test = &new Sieve_Test();
            if ($filter['combine'] == INGO_STORAGE_COMBINE_ANY) {
                $test = &new Sieve_Test_Anyof();
            } else {
                $test = &new Sieve_Test_Allof();
            }

            foreach ($filter['conditions'] as $condition) {
                $tmp = '';
                switch ($condition['match']) {
                    case 'less than':
                        $tmp = &new Sieve_Test_Size(array('comparison' => ':under', 'size' => $condition['value']));
                        $test->addTest($tmp);
                        break;

                    case 'greater than':
                        $tmp = &new Sieve_Test_Size(array('comparison' => ':over', 'size' => $condition['value']));
                        $test->addTest($tmp);
                        break;

                    case 'exists':
                        $tmp = &new Sieve_Test_Exists(array('headers' => $condition['field']));
                        $test->addTest($tmp);
                        break;

                    case 'not exists':
                        $tmp = &new Sieve_Test_Exists(array('headers' => $condition['field']));
                        $test->addTest(new Sieve_Test_Not($tmp));
                        break;

                    case 'contains':
                        $comparator = ($condition['case']) ? 'i;octet' : 'i;ascii-casemap';
                        $vals = array('headers' => $condition['field'],
                                      'match-type' => ':contains',
                                      'strings' => $condition['value'],
                                      'comparator' => $comparator);
                        $tmp = &new Sieve_Test_Header($vals);
                        $test->addTest($tmp);
                        break;

                    case 'not contain':
                        $comparator = ($condition['case']) ? 'i;octet' : 'i;ascii-casemap';
                        $vals = array('headers' => $condition['field'],
                                      'match-type' => ':contains',
                                      'strings' => $condition['value'],
                                      'comparator' => $comparator);
                        $tmp = &new Sieve_Test_Header($vals);
                        $test->addTest(new Sieve_Test_Not($tmp));
                        break;

                    case 'is':
                        $comparator = ($condition['case']) ? 'i;octet' : 'i;ascii-casemap';
                        $vals = array('headers' => $condition['field'],
                                      'match-type' => ':is',
                                      'strings' => $condition['value'],
                                      'comparator' => $comparator);
                        $tmp = &new Sieve_Test_Header($vals);
                        $test->addTest($tmp);
                        break;

                    case 'not is':
                        $comparator = ($condition['case']) ? 'i;octet' : 'i;ascii-casemap';
                        $vals = array('headers' => $condition['field'],
                                      'match-type' => ':is',
                                      'strings' => $condition['value'],
                                      'comparator' => $comparator);
                        $tmp = &new Sieve_Test_Header($vals);
                        $test->addTest(new Sieve_Test_Not($tmp));
                        break;

                    case 'begins with':
                        $comparator = ($condition['case']) ? 'i;octet' : 'i;ascii-casemap';
                        $vals = array('headers' => $condition['field'],
                                      'match-type' => ':matches',
                                      'strings' => $condition['value'] . '*',
                                      'comparator' => $comparator);
                        $tmp = &new Sieve_Test_Header($vals);
                        $test->addTest($tmp);
                        break;

                    case 'not begins with':
                        $comparator = ($condition['case']) ? 'i;octet' : 'i;ascii-casemap';
                        $vals = array('headers' => $condition['field'],
                                      'match-type' => ':matches',
                                      'strings' => $condition['value'] . '*',
                                      'comparator' => $comparator);
                        $tmp = &new Sieve_Test_Header($vals);
                        $test->addTest(new Sieve_Test_Not($tmp));
                        break;

                    case 'ends with':
                        $comparator = ($condition['case']) ? 'i;octet' : 'i;ascii-casemap';
                        $vals = array('headers' => $condition['field'],
                                      'match-type' => ':matches',
                                      'strings' => '*' . $condition['value'],
                                      'comparator' => $comparator);
                        $tmp = &new Sieve_Test_Header($vals);
                        $test->addTest($tmp);
                        break;

                    case 'not ends with':
                        $comparator = ($condition['case']) ? 'i;octet' : 'i;ascii-casemap';
                        $vals = array('headers' => $condition['field'],
                                      'match-type' => ':matches',
                                      'strings' => '*' . $condition['value'],
                                      'comparator' => $comparator);
                        $tmp = &new Sieve_Test_Header($vals);
                        $test->addTest(new Sieve_Test_Not($tmp));
                        break;
                }
            }

            $if = &new Sieve_If($test);
            $if->setActions($action);
            $this->addBlock($if);
        }

        return $this->toCode();
    }

}

/**
 * The Sieve_If:: This class represents a Sieve If Statement
 *
 * @author  Mike Cochrane <mike@graftonhall.co.nz>
 * @since   Ingo 0.1
 * @package ingo
 */
class Sieve_If {

    var $_test;
    var $_actions = array();
    var $_elsifs = array();
    var $_else;

    function Sieve_If($test = null)
    {
        if (is_null($test)) {
            $this->_test = new Sieve_Test_False();
        } else {
            $this->_test = $test;
        }

        $this->_actions[] = new Sieve_Action_Keep();
        $this->_else = new Sieve_Else();
    }

    function getTest()
    {
        return $this->_test;
    }

    function setTest($test)
    {
        $this->_test = $test;
    }

    function getActions()
    {
        return $this->_actions;
    }

    function setActions($actions)
    {
        $this->_actions = $actions;
    }

    function getElsifs()
    {
        return $this->_elsifs;
    }

    function setElsifs($elsifs)
    {
        $this->_elsifs = $elsifs;
    }

    function addElsif($elsif)
    {
        $this->_elsifs[] = $elsif;
    }

    function getElse()
    {
        return $this->_else;
    }

    function setElse($else)
    {
        $this->_else = $else;
    }

    function toCode()
    {
        $code  = 'if ';
        $code .= $this->_test->toCode();
        $code .= " { \n";
        foreach ($this->_actions as $action) {
            $code .= '    ' . $action->toCode() . "\n";
        }
        $code .= "} ";

        foreach ($this->_elsifs as $elsif) {
            $code .= $elsif->toCode();
        }

        $code .= $this->_else->toCode();

        return $code . "\n";
    }

    function check()
    {
        $res = $this->_test->check();
        if ($res !== true) {
            return $res;
        }

        foreach ($this->_elsifs as $elsif) {
            $res = $elsif->check();
            if ($res !== true) {
                return $res;
            }
        }

        $res = $this->_else->check();
        if ($res !== true) {
            return $res;
        }

        foreach ($this->_actions as $action) {
            $res = $action->check();
            if ($res !== true) {
                return $res;
            }
        }

        return true;
    }

    function requires()
    {
        $requires = array();

        foreach ($this->_actions as $action) {
            $requires = array_merge($requires, $action->requires());
        }

        return $requires;
    }

}

/**
 * The Sieve_Else:: This class represents a Sieve Else Statement
 *
 * @author  Mike Cochrane <mike@graftonhall.co.nz>
 * @since   Ingo 0.1
 * @package ingo
 */
class Sieve_Else {

    var $_actions = array();

    function Sieve_Else($actions = null)
    {
        if (is_array($actions)) {
            $this->_actions = $acitons;
        } elseif (!is_null($actions)) {
            $this->_actions[] = $actions;
        }
    }

    function toCode()
    {
        if (count($this->_actions) == 0) {
            return '';
        }

        $code  = 'else';
        $code .= " { \n";
        foreach ($this->_actions as $action) {
            $code .= '    ' . $action->toCode() . "\n";
        }
        $code .= "} ";

        return $code;
    }

    function setActions($actions)
    {
        $this->_actions = $actions;
    }

    function getActions()
    {
        return $this->_actions;
    }

    function check() {
        foreach ($this->_actions as $action) {
            $res = $action->check();
            if ($res !== true) {
                return $res;
            }
        }

        return true;
    }

    function requires()
    {
        $requires = array();

        foreach ($this->_actions as $action) {
            $requires = array_merge($requires, $action->requires());
        }

        return $requires;
    }

}

/**
 * The Sieve_Elsif:: This class represents a Sieve Elsif Statement
 *
 * @author  Mike Cochrane <mike@graftonhall.co.nz>
 * @since   Ingo 0.1
 * @package ingo
 */
class Sieve_Elsif {

    var $_test;
    var $_actions = array();

    function Sieve_Elsif($test = null) {
        if (is_null($test)) {
            $this->_test = new Sieve_Test_False();
        } else {
            $this->_test = $test;
        }
        $this->_actions[] = new Sieve_Action_Keep();
    }

    function getTest()
    {
        return $this->_test;
    }

    function setTest($test)
    {
        $this->_test = $test;
    }

    function getActions()
    {
        return $this->_actions;
    }

    function setActions($actions)
    {
        $this->_actions = $actions;
    }

    function toCode()
    {
        $code  = 'elsif ';
        $code .= $this->_test->toCode();
        $code .= " { \n";
        foreach ($this->_actions as $action) {
            $code .= '    ' . $action->toCode() . "\n";
        }
        $code .= "} ";

        return $code;
    }

    function check() {
        $res = $this->_test->check();
        if ($res !== true) {
            return $res;
        }

        foreach ($this->_actions as $action) {
            $res = $action->check();
            if ($res !== true) {
                return $res;
            }
        }

        return true;
    }

    function requires()
    {
        $requires = array();

        foreach ($this->_actions as $action) {
            $requires = array_merge($requires, $action->requires());
        }

        return $requires;
    }

}

/**
 * The Sieve_Test:: This class represents a Sieve Test. A test is a piece
 * of code that evaluates to true or false.
 *
 * @author  Mike Cochrane <mike@graftonhall.co.nz>
 * @since   Ingo 0.1
 * @package ingo
 */
class Sieve_Test {

    function toCode()
    {
        return 'toCode() Function Not Implemented in class ' . get_class($this);
    }

    function check()
    {
        return 'check() Function Not Implemented in class ' . get_class($this);
    }

    function requires()
    {
        return array();
    }

}

/**
 * The Sieve_Test_True:: This class represents a True test.
 *
 * @author  Mike Cochrane <mike@graftonhall.co.nz>
 * @since   Ingo 0.1
 * @package ingo
 */
class Sieve_Test_True extends Sieve_Test {

    function toCode()
    {
        return 'true';
    }

    function check()
    {
        return true;
    }

}

/**
 * The Sieve_Test_False:: This class represents a False test.
 *
 * @author  Mike Cochrane <mike@graftonhall.co.nz>
 * @since   Ingo 0.1
 * @package ingo
 */
class Sieve_Test_False extends Sieve_Test {

    function toCode()
    {
        return 'false';
    }

    function check()
    {
        return true;
    }

}

/**
 * The Sieve_Test_Allof:: This class represents a Allof test structure.
 * Equivalent to a logical AND of all the tests it contains
 *
 * @author  Mike Cochrane <mike@graftonhall.co.nz>
 * @since   Ingo 0.1
 * @package ingo
 */
class Sieve_Test_Allof extends Sieve_Test {

    var $_tests = array();

    function Sieve_Test_Allof($test = null)
    {
        if (is_array($test)) {
            $this->_tests = $test;
        } elseif (!is_null($test)) {
            $this->_tests[] = $test;
        }
    }

    function toCode()
    {
        $code = '';
        if (count($this->_tests) > 1) {
            $testlist = '';
            foreach ($this->_tests as $test) {
                $testlist .= (empty($testlist)) ? '' : ', ';
                $testlist .= trim($test->toCode());
            }

            $code = "allof ( $testlist )";
        } elseif (count($this->_tests) == 1) {
            $code = $this->_tests[0]->toCode();
        } else {
            return 'false';
        }
        return $code;
    }

    function check() {
        foreach ($this->_tests as $test) {
            $res = $test->check();
            if ($res !== true) {
                return $res;
            }
        }

        return true;
    }

    function addTest($test)
    {
        $this->_tests[] = $test;
    }

    function getTests()
    {
        return $this->_tests;
    }

}

/**
 * The Sieve_Test_Anyof:: This class represents a Anyof test structure.
 * Equivalent to a logical OR of all the tests it contains
 *
 * @author  Mike Cochrane <mike@graftonhall.co.nz>
 * @since   Ingo 0.1
 * @package ingo
 */
class Sieve_Test_Anyof extends Sieve_Test {

    var $_tests = array();

    function Sieve_Test_Anyof($test = null)
    {
        if (is_array($test)) {
            $this->_tests = $test;
        } elseif (!is_null($test)) {
            $this->_tests[] = $test;
        }
    }

    function toCode()
    {
        $testlist = '';
        if (count($this->_tests) > 1) {
            $testlist = '';
            foreach ($this->_tests as $test) {
                $testlist .= (empty($testlist)) ? '' : ', ';
                $testlist .= trim($test->toCode());
            }

            $code = "anyof ( $testlist )";
        } elseif (count($this->_tests) == 1) {
            $code = $this->_tests[0]->toCode();
        } else {
            return 'false';
        }
        return $code;
    }

    function addTest($test)
    {
        $this->_tests[] = $test;
    }

    function getTests()
    {
        return $this->_tests;
    }

    function check() {
        foreach ($this->_tests as $test) {
            $res = $test->check();
            if ($res !== true) {
                return $res;
            }
        }

        return true;
    }

}

/**
 * The Sieve_Test_Size:: This class represents a message size test.
 *
 * @author  Mike Cochrane <mike@graftonhall.co.nz>
 * @since   Ingo 0.1
 * @package ingo
 */
class Sieve_Test_Size extends Sieve_Test {

    var $_vars = array();

    function Sieve_Test_Size($vars = array())
    {
        $this->_vars['comparison'] = (isset($vars['comparison'])) ? $vars['comparison'] : '';
        $this->_vars['size'] = (isset($vars['size'])) ? $vars['size'] : '';
    }

    function toCode()
    {
        $code  = '';
        $code .= 'size ';
        $code .= $this->_vars['comparison'];
        $code .= ' ' . $this->_vars['size'];

        return $code;
    }

    function check()
    {
        if (!(isset($this->_vars['comparison'])
                && isset($this->_vars['size']))) {
            return false;
        }

        return true;
    }

}

/**
 * The Sieve_Test_Not:: This class represents the inverse of a given test.
 *
 * @author  Mike Cochrane <mike@graftonhall.co.nz>
 * @since   Ingo 0.1
 * @package ingo
 */
class Sieve_Test_Not extends Sieve_Test {

    var $_test = array();

    function Sieve_Test_Not($test)
    {
        $this->_test = $test;
    }

    function check()
    {
        return $this->_test->check();
    }

    function toCode()
    {
        $code  = 'not ';
        $code .= $this->_test->toCode();
        return $code;
    }

}

/**
 * The Sieve_Test_Exists:: This class represents a test for the existsance of
 * one or more headers in a message.
 *
 * @author  Mike Cochrane <mike@graftonhall.co.nz>
 * @since   Ingo 0.1
 * @package ingo
 */
class Sieve_Test_Exists extends Sieve_Test {

    var $_vars = array();

    function Sieve_Test_Exists($vars = array())
    {
        $this->_vars['headers'] = (isset($vars['headers'])) ? $vars['headers'] : '';
    }

    function check()
    {
        $headers = preg_split('(\r\n|\n|\r)', $this->_vars['headers']);
        if (count($headers) < 1) {
            return _("Not headers specified");
        }

        return true;
    }

    function toCode()
    {
        $code  = '';
        $code .= 'exists ';
        $headers = preg_split('(\r\n|\n|\r)', $this->_vars['headers']);
        if (count($headers) > 1) {
            $code .= "[";
            $headerstr = '';
            foreach ($headers as $header) {
                $headerstr .= empty($headerstr) ? '"' : ', "';
                $headerstr .= Ingo_Script_sieve::escapeString($header) . '"';
            }
            $code .= $headerstr . "] ";
        } else if (count($headers) == 1) {
            $code .= '"' . Ingo_Script_sieve::escapeString($headers[0]) . '" ';
        } else {
            return "**error** No Headers Specified";
        }

        return $code;
    }

}

/**
 * The Sieve_Test_Address:: This class represents a test on parts or all of
 * the addresses in the given fields.
 *
 * @author  Mike Cochrane <mike@graftonhall.co.nz>
 * @since   Ingo 0.1
 * @package ingo
 */
class Sieve_Test_Address extends Sieve_Test {

    var $_vars = array();

    function Sieve_Test_Address($vars)
    {
        $this->_vars['headers'] = (isset($vars['headers'])) ? $vars['headers'] : '';
        $this->_vars['comparator'] = (isset($vars['comparator'])) ? $vars['comparator'] : 'i;ascii-casemap';
        $this->_vars['match-type'] = (isset($vars['match-type'])) ? $vars['match-type'] : ':is';
        $this->_vars['address-part'] = (isset($vars['address-part'])) ? $vars['address-part'] : ':all';
        $this->_vars['addresses'] = (isset($vars['addresses'])) ? $vars['addresses'] : '';
    }

    function check()
    {
        $headers = preg_split('(\r\n|\n|\r)', $this->_vars['headers']);
        if (count($headers) < 1) {
            return false;
        }

        $addresses = preg_split('(\r\n|\n|\r)', $this->_vars['addresses']);
        if (count($addresses) < 1) {
            return false;
        }

        return true;
    }

    function toCode()
    {
        $code  = '';
        $code .= 'address ';

        $code .= $this->_vars['address-part'] . ' ';

        $code .= ':comparator "' . $this->_vars['comparator'] . '" ';

        $code .= $this->_vars['match-type'] . ' ';

        $headers = preg_split('(\r\n|\n|\r)', $this->_vars['headers']);
        if (count($headers) > 1) {
            $code .= "[";
            $headerstr = '';
            foreach ($headers as $header) {
                $header = trim($header);
                if (!empty($header)) {
                    $headerstr .= empty($headerstr) ? '"' : ', "';
                    $headerstr .= Ingo_Script_sieve::escapeString($header) . '"';
                }
            }
            $code .= $headerstr . "] ";
        } else if (count($headers) == 1) {
            $code .= '"' . Ingo_Script_sieve::escapeString($headers[0]) . '" ';
        } else {
            return "No Headers Specified";
        }

        $addresses = preg_split('(\r\n|\n|\r)', $this->_vars['addresses']);
        if (count($addresses) > 1) {
            $code .= "[";
            $addressstr = '';
            foreach ($addresses as $addr) {
                $addr = trim($addr);
                if (!empty($addr)) {
                    $addressstr .= empty($addressstr) ? '"' : ', "';
                    $addressstr .= Ingo_Script_sieve::escapeString($addr) . '"';
                }
            }
            $code .= $addressstr . "] ";
        } else if (count($addresses) == 1) {
            $code .= '"' . Ingo_Script_sieve::escapeString($addresses[0]) . '" ';
        } else {
            return "No Addresses Specified";
        }

        return $code;
    }

}

/**
 * The Sieve_Test_Header:: This class represents a test on the contents of
 * one or more headers in a message.
 *
 * @author  Mike Cochrane <mike@graftonhall.co.nz>
 * @since   Ingo 0.1
 * @package ingo
 */
class Sieve_Test_Header extends Sieve_Test {

    var $_vars = array();

    function Sieve_Test_Header($vars = array())
    {
        $this->_vars['headers'] = (isset($vars['headers'])) ? $vars['headers'] : 'Subject';
        $this->_vars['comparator'] = (isset($vars['comparator'])) ? $vars['comparator'] : 'i;ascii-casemap';
        $this->_vars['match-type'] = (isset($vars['match-type'])) ? $vars['match-type'] : ':is';
        $this->_vars['strings'] = (isset($vars['strings'])) ? $vars['strings'] : '';
    }

    function check()
    {
        $headers = preg_split('(\r\n|\n|\r)', $this->_vars['headers']);
        if (count($headers) < 1) {
            return false;
        }

        $strings = preg_split('(\r\n|\n|\r)', $this->_vars['strings']);
        if (count($strings) < 1) {
            return false;
        }

        return true;
    }

    function toCode()
    {
        $code  = '';
        $code .= 'header ';

        $code .= ':comparator "' . $this->_vars['comparator'] . '" ';

        $code .= $this->_vars['match-type'] . ' ';

        $headers = preg_split('(\r\n|\n|\r)', $this->_vars['headers']);
        if (count($headers) > 1) {
            $code .= "[";
            $headerstr = '';
            foreach ($headers as $header) {
                $headerstr .= empty($headerstr) ? '"' : ', "';
                $headerstr .= Ingo_Script_sieve::escapeString($header) . '"';
            }
            $code .= $headerstr . "] ";
        } else if (count($headers) == 1) {
            $code .= '"' . $headers[0] . '" ';
        } else {
            return _("No Headers Specified");
        }

        $strings = preg_split('(\r\n|\n|\r)', $this->_vars['strings']);
        if (count($strings) > 1) {
            $code .= "[";
            $stringlist = '';
            foreach ($strings as $str) {
                $stringlist .= empty($stringlist) ? '"' : ', "';
                $stringlist .= Ingo_Script_sieve::escapeString($str) . '"';
            }
            $code .= $stringlist . "] ";
        } else if (count($strings) == 1) {
            $code .= '"' . Ingo_Script_sieve::escapeString($strings[0]) . '" ';
        } else {
            return _("No Strings Specified");
        }

        return $code;
    }

}

/**
 * A Comment. This and Sieve_If should really extends a Sieve_Block eventually.
 *
 * @author  Mike Cochrane <mike@graftonhall.co.nz>
 * @since   Ingo 0.1
 * @package ingo
 */
class Sieve_Comment {

    var $_comment;

    function Sieve_Comment($comment)
    {
        $this->_comment = $comment;
    }

    function toCode()
    {
        $code = '';
        $lines = preg_split('(\r\n|\n|\r)', $this->_comment);
        foreach ($lines as $line) {
            $line = trim($line);
            if (!empty($line)) {
                $code .= (empty($code)) ? '' : "\n";
                $code .= '# ' . $line;
            }
        }
        return $code;
    }

    function check()
    {
        return true;
    }

    function requires()
    {
        return array();
    }

}

/**
 * The Sieve_Action:: This class represents an action in a Sieve script.
 * An action is anything that has a side effect eg: discard, redirect.
 *
 * @author  Mike Cochrane <mike@graftonhall.co.nz>
 * @since   Ingo 0.1
 * @package ingo
 */
class Sieve_Action {

    function toCode()
    {
        return 'toCode() Function Not Implemented in class ' . get_class($this) ;
    }

    function toString()
    {
        return $this->toCode();
    }

    function check()
    {
        return 'check() Function Not Implemented in class ' . get_class($this) ;
    }

    function requires()
    {
        return array();
    }

}

/**
 * The Sieve_Action_Redirect:: This class represents a redirect action.
 *
 * @author  Mike Cochrane <mike@graftonhall.co.nz>
 * @since   Ingo 0.1
 * @package ingo
 */
class Sieve_Action_Redirect extends Sieve_Action {

    var $_vars = array();

    function Sieve_Action_Redirect($vars = array())
    {
        $this->_vars['address'] = (isset($vars['address'])) ? $vars['address'] : '';
    }

    function toCode($depth = 0) {
        $code  = str_repeat(' ', $depth * 4);
        $code .= 'redirect ';
        $code .= '"' .  Ingo_Script_sieve::escapeString($this->_vars['address']) . '";';
        return $code;
    }

    function check()
    {
        if (empty($this->_vars['address'])) {
            return _("Missing Address to Redirect message to");
        }

        return true;
    }

}

/**
 * The Sieve_Action_Reject:: class represents a reject action.
 *
 * @author  Mike Cochrane <mike@graftonhall.co.nz>
 * @since   Ingo 0.1
 * @package ingo
 */
class Sieve_Action_Reject extends Sieve_Action {

    var $_vars = array();

    function Sieve_Action_Reject($vars = array())
    {
        $this->_vars['reason'] = (isset($vars['reason'])) ? $vars['reason'] : '';
    }

    function toCode() {
        $code  = '';
        $code .= 'reject ';
        $code .= '"' .  Ingo_Script_sieve::escapeString($this->_vars['reason']) . '";';
        return $code;
    }

    function check()
    {
        if (empty($this->_vars['reason'])) {
            return _("Missing Reason for Reject");
        }

        return true;
    }

    function requires()
    {
        return array('reject');
    }

}

/**
 * The Sieve_Action_Keep:: class represents a keep action.
 *
 * @author  Mike Cochrane <mike@graftonhall.co.nz>
 * @since   Ingo 0.1
 * @package ingo
 */
class Sieve_Action_Keep extends Sieve_Action {

    function toCode()
    {
        $code = 'keep;';
        return $code;
    }

    function check()
    {
        return true;
    }

}

/**
 * The Sieve_Action_Discard:: class represents a discard action.
 *
 * @author  Mike Cochrane <mike@graftonhall.co.nz>
 * @since   Ingo 0.1
 * @package ingo
 */
class Sieve_Action_Discard extends Sieve_Action {

    function toCode()
    {
        $code = 'discard;';
        return $code;
    }

    function check()
    {
        return true;
    }

}

/**
 * The Sieve_Action_Stop:: class represents a stop action.
 *
 * @author  Mike Cochrane <mike@graftonhall.co.nz>
 * @since   Ingo 0.1
 * @package ingo
 */
class Sieve_Action_Stop extends Sieve_Action {

    function toCode()
    {
        $code = 'stop;';
        return $code;
    }

    function check()
    {
        return true;
    }

}

/**
 * The Sieve_Action_Fileinto:: class represents a fileinto action.
 *
 * @author  Mike Cochrane <mike@graftonhall.co.nz>
 * @since   Ingo 0.1
 * @package ingo
 */
class Sieve_Action_Fileinto extends Sieve_Action {

    var $_vars = array();

    function Sieve_Action_Fileinto($vars = array())
    {
        $this->_vars['folder'] = (isset($vars['folder'])) ? $vars['folder'] : '';
    }

    function toCode() {
        $code = 'fileinto "' .  Ingo_Script_sieve::escapeString($this->_vars['folder']) . '";';
        return $code;
    }

    function check()
    {
        if (empty($this->_vars['folder'])) {
            return _("Mailbox Missing to deliver the message into");
        }

        return true;
    }

    function requires()
    {
        return array('fileinto');
    }

}

/**
 * The Sieve_Action_Vacation:: class represents a vacation action.
 *
 * @author  Mike Cochrane <mike@graftonhall.co.nz>
 * @since   Ingo 0.1
 * @package ingo
 */
class Sieve_Action_Vacation extends Sieve_Action {

    var $_vars = array();

    function Sieve_Action_Vacation($vars = array())
    {
        $this->_vars['days'] = (isset($vars['days'])) ? intval($vars['days']) : 7;
        $this->_vars['addresses'] = (isset($vars['addresses'])) ? $vars['addresses'] : '';
        $this->_vars['subject'] = (isset($vars['subject'])) ? $vars['subject'] : '';
        $this->_vars['reason'] = (isset($vars['reason'])) ? $vars['reason'] : '';
    }

    function toCode() {
        $code  = '';
        $code .= 'vacation ';
        $code .= ':days ' . $this->_vars['days'] . ' ';

        $addresses = preg_split('(\r\n|\n|\r)', $this->_vars['addresses']);

        $stringlist = '';
        if (count($addresses) > 1) {
            foreach ($addresses as $address) {
                $address = trim($address);
                if (!empty($address)) {
                    $stringlist .= empty($stringlist) ? '"' : ', "';
                    $stringlist .= Ingo_Script_sieve::escapeString($address) . '"';
                }
            }
            $stringlist = "[" . $stringlist . "] ";
        } else if (count($addresses) == 1) {
            $stringlist = '"' . Ingo_Script_sieve::escapeString($addresses[0]) . '" ';
        }

        if (!empty($stringlist)) {
            $code .= ':addresses ' . $stringlist;
        }

        if (!empty($this->_vars['subject'])) {
            $code .= ':subject "' . Ingo_Script_sieve::escapeString($this->_vars['subject']) . '" ';
        }
        $code .= '"' .  Ingo_Script_sieve::escapeString($this->_vars['reason']) . '";';
        return $code;
    }

    function check()
    {
        if (empty($this->_vars['reason'])) {
            return _("Missing reason in vacation.");
        }

        return true;
    }

    function requires()
    {
        return array('vacation');
    }

}

/**
 * The Sieve_Action_Addflag:: class represents an add flag action.
 *
 * @author  Mike Cochrane <mike@graftonhall.co.nz>
 * @since   Ingo 0.1
 * @package ingo
 */
class Sieve_Action_Addflag extends Sieve_Action {

    var $_vars = array();

    function Sieve_Action_Addflag($vars = array())
    {
        $this->_vars['flags'] = (isset($vars['flags'])) ? $vars['flags'] : '';
    }

    function toCode() {
        $code  = '';
        $code .= 'addflag ';

        if (is_array($this->_vars['flags']) && count($this->_vars['flags']) > 1) {
            $stringlist = '';
            foreach ($this->_vars['flags'] as $flag) {
                $flag = trim($flag);
                if (!empty($flag)) {
                    $stringlist .= empty($stringlist) ? '"' : ', "';
                    $stringlist .= Ingo_Script_sieve::escapeString($flag) . '"';
                }
            }
            $stringlist = '[' . $stringlist . ']';
            $code .= $stringlist . ';';
        } elseif (is_array($this->_vars['flags']) && count($this->_vars['flags']) == 1) {
            $code .= '"' . Ingo_Script_sieve::escapeString($this->_vars['flags'][0]) . '";';
        } else {
            return '';
        }
        return $code;
    }

    function check()
    {
        return true;
    }

    function requires()
    {
        return array('imapflags');
    }

}

/**
 * The Sieve_Action_Removeflag:: class represents a remove flag
 * action.
 *
 * @author  Mike Cochrane <mike@graftonhall.co.nz>
 * @since   Ingo 0.1
 * @package ingo
 */
class Sieve_Action_Removeflag extends Sieve_Action {

    var $_vars = array();

    function Sieve_Action_Removeflag($vars = array())
    {
        $this->_vars['flags'] = (isset($vars['flags'])) ? $vars['flags'] : '';
    }

    function toCode() {
        $code  = '';
        $code .= 'removeflag ';

        if (is_array($this->_vars['flags']) && count($this->_vars['flags']) > 1) {
            $stringlist = '';
            foreach ($this->_vars['flags'] as $flag) {
                $flag = trim($flag);
                if (!empty($flag)) {
                    $stringlist .= empty($stringlist) ? '"' : ', "';
                    $stringlist .= Ingo_Script_sieve::escapeString($flag) . '"';
                }
            }
            $stringlist = '[' . $stringlist . ']';
            $code .= $stringlist . ';';
        } elseif (is_array($this->_vars['flags']) && count($this->_vars['flags']) == 1) {
            $code .= '"' . Ingo_Script_sieve::escapeString($this->_vars['flags'][0]) . '";';
        } else {
            return '';
        }
        return $code;
    }

    function check()
    {
        return true;
    }

    function requires()
    {
        return array('imapflags');
    }

}
