<?php

// Ingo_Storage:: 'combine' constants
/** @const INGO_STORAGE_COMBINE_ALL */
define('INGO_STORAGE_COMBINE_ALL', 1);

/** @const INGO_STORAGE_COMBINE_ANY */
define('INGO_STORAGE_COMBINE_ANY', 2);

// Ingo_Storage:: 'action' constants
/** @const INGO_STORAGE_ACTION_KEEP */
define('INGO_STORAGE_ACTION_KEEP', 1);

/** @const INGO_STORAGE_ACTION_MOVE */
define('INGO_STORAGE_ACTION_MOVE', 2);

/** @const INGO_STORAGE_ACTION_DISCARD */
define('INGO_STORAGE_ACTION_DISCARD', 3);

/** @const INGO_STORAGE_ACTION_REDIRECT */
define('INGO_STORAGE_ACTION_REDIRECT', 4);

/** @const INGO_STORAGE_ACTION_REDIRECTKEEP */
define('INGO_STORAGE_ACTION_REDIRECTKEEP', 5);

/** @const INGO_STORAGE_ACTION_REJECT */
define('INGO_STORAGE_ACTION_REJECT', 6);

/** @const INGO_STORAGE_ACTION_BLACKLIST */
define('INGO_STORAGE_ACTION_BLACKLIST', 7);

/** @const INGO_STORAGE_ACTION_VACATION */
define('INGO_STORAGE_ACTION_VACATION', 8);

/** @const INGO_STORAGE_ACTION_WHITELIST */
define('INGO_STORAGE_ACTION_WHITELIST', 9);

/**
 * Ingo_Storage:: defines an API to store the various filter rules.
 *
 * All storage drivers should handle the following fields:
 * 'rules' (array)             -- The listing of filter rules.

 * 'blacklist' (string)        -- The listing of blacklisted addresses.
 * 'blacklist_folder' (string) -- The blacklist target folder.
 * 'whitelist' (string)        -- The listing of whitelisted addresses.
 * 'vacation' (array)          -- The vaction settings.
 *
 * $Horde: ingo/lib/Storage.php,v 1.6.2.2 2003/05/07 15:13:20 bjn Exp $
 *
 * @author  Michael Slusarz <slusarz@bigworm.colorado.edu>
 * @version $Revision: 1.6.2.2 $
 * @since   Ingo 0.1
 * @package ingo
 */
class Ingo_Storage {

    /**
     * Driver specific parameters
     *
     * @var array $_params
     */
    var $_params;

    /**
     * Attempts to return a concrete Ingo_Storage instance based on $driver.
     *
     * @access public
     *
     * @param string $driver          The type of concrete Ingo_Storage
     *                                subclass to return.  This is based on
     *                                the storage driver ($driver).  The code
     *                                is dynamically included.
     * @param array optional $params  A hash containing any additional
     *                                configuration or connection parameters
     *                                a subclass might need.
     *
     * @return mixed  The newly created concrete Ingo_Storage instance, or
     *                false on an error.
     */
    function &factory($driver, $params = array())
    {
        $driver = basename($driver);
        require_once dirname(__FILE__) . '/Storage/' . $driver . '.php';
        $class = 'Ingo_Storage_' . $driver;
        if (class_exists($class)) {
            return new $class($params);
        } else {
            return false;
        }
    }

    /**
     * Attempts to return a reference to a concrete Ingo_Storage instance
     * based on $driver.  It will only create a new instance if no
     * Ingo_Storage instance with the same parameters currently exists.
     *
     * This should be used if multiple storage sources are required.
     *
     * This method must be invoked as: $var = &Ingo_Storage::singleton();
     *
     * @param string $driver          The type of concrete Ingo_Storage
     *                                subclass to return.  This is based on
     *                                the storage driver ($driver).  The code
     *                                is dynamically included.
     * @param array optional $params  A hash containing any additional
     *                                configuration or connection parameters
     *                                a subclass might need.
     *
     * @return mixed  The created concrete Ingo_Storage instance, or false
     *                on error.
     */
    function &singleton($driver, $params = array())
    {
        static $instances;

        if (!isset($instances)) {
            $instances = array();
        }

        $signature = serialize(array($driver, $params));
        if (!isset($instances[$signature])) {
            $instances[$signature] = &Ingo_Storage::factory($driver, $params);
        }

        return $instances[$signature];
    }

    /**
     * Retrieve the specified data.
     *
     * @access public
     *
     * @param string $field  The field name of the desired data.
     *                       See above for the available fields.
     *
     * @return array  The specified data.
     */
    function retrieve($field)
    {
        return false;
    }

    /**
     * Store the specified data.
     *
     * @access public
     *
     * @param string $field  The field name to store the desired data.
     *                       See above for the available fields.
     * @param mixed $data    The data to store.
     *
     * @return boolean  True on success.
     */
    function store($field, $data)
    {
        return false;
    }

    /**
     * Is this field locked?
     *
     * @access public
     *
     * @param string $field  The field name to store the desired data.
     *                       See above for the available fields.
     *
     * @return boolean  Whether the field is locked or not.
     */
    function isLocked($field)
    {
        return false;
    }

    /**
     * Prepare a list of addresses for storage.
     *
     * @access private
     *
     * @param array $addr  The list of addresses.
     *
     * @return array  The list of addresses, prepared for storage.
     */
    function _prepareAddressList($addr)
    {
        /* Remove any extra space in the address and make it lowercase. */
        //$addr = array_map('trim', $addr);
        //$addr = array_map(array('String', 'lower'), $addr);

        /* Remove duplicate entries. */
        $addr = array_unique($addr);

        /* Sort the list. */
        usort($addr, array($this, '_sortAddressList'));

        return $addr;
    }

    /**
     * Function used by usort() to sort an address list.
     *
     * @access private
     *
     * @param string $a  Address #1.
     * @param string $b  Address #2.
     *
     * @return integer  -1, 0, or 1.
     */
    function _sortAddressList($a, $b)
    {
        $a = explode('@', $a);
        $b = explode('@', $b);

        /* One of the addresses doesn't have a host name. */
        if (empty($a[0])) {
            array_shift($a);
        }
        if (empty($b[0])) {
            array_shift($b);
        }
        if (count($a) != count($b)) {
            return (count($a) > count($b));
        }

        /* The addresses have different hostname or not hostname and
           different mailbox names. */
        if ($a[(count($a) - 1)] != $b[(count($b) - 1)]) {
            return strcmp($a[(count($a) - 1)], $b[(count($b) - 1)]);
        }

        /* Compare mailbox names. */
        return strcmp($a[0], $b[0]);
    }

}
